/* ----------------------------------------------------------------- */
/* --- Koordinierung von Weihnachtsmann, Elfen und Rentieren mit --- */
/* --- mit Hilfe von Semaphoren -------------------------------------*/
/* ----------------------------------------------------------------- */

/* --- Der Weihnachtsmann schlft in seinem Haus am Nordpol.     --- */
/* --- Wenn die Elfen Schwierigkeiten bei der Produktion des     --- */
/* --- Spielzeug haben, fragen sie den Weihnachtsmann um Rat.    --- */
/* --- Es mssen aber immer drei Elfen zusammenkommen, damit der --- */
/* --- Weihnachtsmann nicht dauernd gestrt wird. Wenn noch mehr --- */
/* --- Elfen Probleme haben, mssen sie warten, bis die anderen  --- */
/* --- wieder da sind.                                           --- */
/* --- Die Rentiere verbringen ihre Ferien im sonnigen Sden.    --- */
/* --- Kurz vor Weihnachten kommen sie zurck an den Nordpol.    --- */
/* --- Wenn alle neuen Rentiere da sind, ist Weihnachten, un der --- */
/* --- Weihnachtsmann spannt die Rentiere vor den Schlitten, um  --- */
/* --- seiner Arbeit nachzugehen (die Rentiere kommen immer erst --- */
/* --- im letzten Moment zurck). Die Rentiere haben immer       --- */
/* --- Vorrang vor wartenden Elfen, da Weihnachten sonst         --- */
/* --- verschoben wrde. Nach Weihnachten kehren sie zurck in   --- */
/* --- ins Warme.                                                --- */

#include "xmas.h"

//ID der Semaphore
//ID des shared memory
int semid;
int shmid;

struct sembuf sops[SEM_NUM];
union semun args;

//shared memory pointer
shared_t *shmPtr;

//Name des Rentiers
char *name;

//Semaphor freigeben
void up(int semnum)
{
    sops[0].sem_num = semnum;
    sops[0].sem_op = 1;
    sops[0].sem_flg = 0;

    if( 0 > semop(semid, &(sops[0]), 1))
    {
	perror("reindeer: problem with semop()");
	exit(1);
    }
}

//Semaphor anfordern
void down(int semnum)
{
    sops[0].sem_num = semnum;
    sops[0].sem_op = -1;
    sops[0].sem_flg = 0;

    if( 0 > semop(semid, &(sops[0]), 1))
    {
	perror("reindeer: problem with semop()");
	exit(1);
    }
}

//Signalhandler
void sighandler(int sig)
{
    printf("Signal %d caught for reindeer.\n", sig);
    
    //Semaphore freigeben
    if(0 > semctl(semid, SEM_NUM, IPC_RMID))
    {
	perror("reindeer: problem with semctl()");
	exit(1);
    }

    //Shared Memory ausblenden
    if(0 > shmdt(shmPtr))
    {
	perror("reindeer: problem with shmdt()");
	exit(1);
    }

    //Shared Memory freigeben
    if(0 > shmctl(shmid, IPC_RMID, NULL))
    {
	perror("reindeer: problem with shmctl()");
	exit(1);
    }

    free(name);

    exit(0);
}

int main(int argc, char* argv[])
{
    //Schlafen
    int duration;

    //ein Argument: Name des Rentiers
    if(argc != 2)
    {
	perror("reindeer: wrong arguments");
	exit(1);
    }

    //Name speichern
    //Achtung: strlen liefert die Anzahl der Zeichen
    //das abschlieende \0 muss auch bercksichtigt werden
    name = malloc((strlen(argv[1]) + 1) * sizeof(char));
    memcpy(name, argv[1], strlen(argv[1]) + 1);

    //Semaphor holen
    if ( 0 > (semid = semget(SEM_KEY,
                             SEM_NUM,
                             (IPC_CREAT | SEM_PERM))) ) {
	perror("reindeer: problem with semget()");
        exit(1);
    }

    //shared memory holen
    if( 0 > (shmid = shmget(SHM_KEY, sizeof(shared_t), (IPC_CREAT | SHM_PERM))))
    {
	perror("reindeer: problem with shmget()");
	exit(1);
    }

    //shared memory binden
    if(0 > (shmPtr = shmat(shmid, 0, 0)))
    {
	perror("reindeer:problem with shmat()");
	exit(1);
    }

    //Zufallszahlen initialisieren
    srandom(time(NULL) + (int)name[0] + (int)name[strlen(name)-1] + getpid());

    //auf Signale hren
    signal(SIGTERM, sighandler);
    signal(SIGINT, sighandler);

    printf("Reindeer %s has finished init phase\n", name);

    //Semphor da, jetzt kann der eigentliche Prozess loslegen
    while(1)
    {
	printf("Reindeer %s enjoys the sun.\n", name);
	
	duration = 10+random()%60;
	sleep(duration);

	printf("Reindeer %s is going to polar circle.\n", name);

	//kann dauern, ja nachdem wo das Rentier gerade
	//Urlaub gemacht hat
	duration = random()%5;
	sleep(duration);

	//beim Weihnachtsmann melden
	down(SEM_MUTEX);
	shmPtr->numberOfReindeers++;
	//der letze muss den Weihnachtsmann rufen
	if(shmPtr->numberOfReindeers == ALL_REINDEERS)
	{
	    printf("Reindeer %s wakes santa.\n", name);
	    up(SEM_MUTEX);
	    down(SEM_BACK);
	    up(SEM_WAKEUP);
	}
	else
	    up(SEM_MUTEX);

	//auf den Weihnachtsabend warten
	down(SEM_WORK);
	
	printf("Reindeer %s pulls santa's sleigh.\n", name);

	//fertig
	down(SEM_WORK_DONE);

	down(SEM_MUTEX);
	shmPtr->numberOfReindeers--;
	if(shmPtr->numberOfReindeers == 0)
	{
	    printf("Reindeer %s is the last to leave.\n", name);
	    up(SEM_BACK);
	}
	up(SEM_MUTEX);

	down(SEM_BACK);
	up(SEM_BACK);

	duration = random()%5;
	sleep(duration);

	printf("Reindeer %s returns to nice sunny place.\n", name);
	//tschss sagen
	up(SEM_BYE);
    }
}
