/* ----------------------------------------------------------------- */
/* --- Koordinierung von Weihnachtsmann, Elfen und Rentieren mit --- */
/* --- mit Hilfe von Semaphoren -------------------------------------*/
/* ----------------------------------------------------------------- */

/* --- Der Weihnachtsmann schlft in seinem Haus am Nordpol.     --- */
/* --- Wenn die Elfen Schwierigkeiten bei der Produktion des     --- */
/* --- Spielzeug haben, fragen sie den Weihnachtsmann um Rat.    --- */
/* --- Es mssen aber immer drei Elfen zusammenkommen, damit der --- */
/* --- Weihnachtsmann nicht dauernd gestrt wird. Wenn noch mehr --- */
/* --- Elfen Probleme haben, mssen sie warten, bis die anderen  --- */
/* --- wieder da sind.                                           --- */
/* --- Die Rentiere verbringen ihre Ferien im sonnigen Sden.    --- */
/* --- Kurz vor Weihnachten kommen sie zurck an den Nordpol.    --- */
/* --- Wenn alle neuen Rentiere da sind, ist Weihnachten, un der --- */
/* --- Weihnachtsmann spannt die Rentiere vor den Schlitten, um  --- */
/* --- seiner Arbeit nachzugehen (die Rentiere kommen immer erst --- */
/* --- im letzten Moment zurck). Die Rentiere haben immer       --- */
/* --- Vorrang vor wartenden Elfen, da Weihnachten sonst         --- */
/* --- verschoben wrde. Nach Weihnachten kehren sie zurck in   --- */
/* --- ins Warme.                                                --- */

#include "xmas.h"

//ID fr die Semaphore
//ID frs shared memory
int semid;
int shmid;

//Semaphor Operationen und Argumente
struct sembuf sops[SEM_NUM];
union semun args;

//Pointer frs shared memory
shared_t *shmPtr;

//Semaphor freigeben
void up(int semnum)
{
    sops[0].sem_num = semnum;
    sops[0].sem_op = 1;
    sops[0].sem_flg = 0;

    if( 0 > semop(semid, &(sops[0]), 1))
    {
	perror("santa: problem with semop()");
	exit(1);
    }
}

//Semaphor anfordern
void down(int semnum)
{
    sops[0].sem_num = semnum;
    sops[0].sem_op = -1;
    sops[0].sem_flg = 0;

    if( 0 > semop(semid, &(sops[0]), 1))
    {
	perror("santa: problem with semop()");
	exit(1);
    }
}

//Semaphor initialisieren
void init(int semid, int semnum, int val)
{
    //Initialisieren mit Wert val
    memset(&args, 0, sizeof(union semun));
    args.val = val;

    if ( 0 > semctl(semid,semnum,SETVAL,args) ) 
    {
        perror("santa: problem with semctl(SEM_NUMBER_REINDEERS)");
        exit(1);
    }
}

//Signalhandler
void sighandler(int sig)
{
    printf("Signal %d caught for santa.\n", sig);
    
    //Semaphore freigeben
    if(0 > semctl(semid, SEM_NUM, IPC_RMID))
    {
	perror("santa: problem with semctl()");
	exit(1);
    }

    //Shared Memory ausblenden
    if(0 > shmdt(shmPtr))
    {
	perror("santa: problem with shmget()");
	exit(1);
    }

    //Shared Memory freigeben
    if(0 > shmctl(shmid, IPC_RMID, NULL))
    {
	perror("santa: problem with shmctl()");
	exit(1);
    }

    exit(0);
}

int main()
{
    //Zhler
    int i;

    //Schlafen
    int duration;

    //Zufallszahlen initialisieren
    srandom(time(NULL) + getpid());

    //auf Signale hren
    signal(SIGTERM, sighandler);
    signal(SIGINT, sighandler);

    //shard memory holen
    if( 0 > (shmid = shmget(SHM_KEY, sizeof(shared_t), (IPC_CREAT | SHM_PERM))))
    {
	perror("santa: problem with shmget()");
	exit(1);
    }

    //shared memory in den lokalen Adressraum abbilden
    if (0 > (shmPtr = shmat(shmid, 0, 0)))
    {
	perror("santa: problem with shmat()");
	exit(1);
    }   

    //Semaphore holen
    if ( 0 > (semid = semget(SEM_KEY,
                             SEM_NUM,
                             (IPC_CREAT | SEM_PERM))) ) {
	perror("santa: problem with semget()");
        exit(1);
    }
    
    //die beiden Mutexe sind 1
    init(semid, SEM_MUTEX, 1);

    //Elfen drfen immer drei, deshalb ist die Kapazitt 3
    init(semid, SEM_CAPACITY, 3);

    //Rentiere mssen alle den Polarkreis verlassen, bevor es wieder
    //losgeht
    //am Anfang muss ich anfangen knnen
    init(semid, SEM_BACK, 1);
    init(semid, SEM_LEFT, 1);

    //alle anderen starten mit 0
    init(semid, SEM_WAKEUP, 0);
    init(semid, SEM_WORK, 0);
    init(semid, SEM_WORK_DONE, 0);
    init(semid, SEM_BYE, 0);
    init(semid, SEM_GO_SANTA, 0);
    init(semid, SEM_ADVISE_GIVEN, 0);
    init(semid, SEM_THANKS, 0);

    //jetzt sollte das shared memory noch initialisiert werden
    //deshalb: Semaphor anfordern
    down(SEM_MUTEX);
    shmPtr->numberOfReindeers = 0;
    shmPtr->numberOfElfes = 0;
    up(SEM_MUTEX);

    printf("Santa has finished init phase.\n");

    //Semphor da, jetzt kann der eigentliche Prozess loslegen
    while(1)
    {
	printf("\nSanta's sleeping.\n");

	//damit man sich das besser angucken kann
	sleep(3);

	//warten, bis jemand den Weihnachtsmann weckt
	down(SEM_WAKEUP);
      
	printf("\nSanta is looking for the reindeers or elfes.\n");

	//sind die Rentiere alle da?
	//Mutex fr die globale Variable anfragen
        down(SEM_MUTEX);

	if(shmPtr->numberOfReindeers == ALL_REINDEERS)
	{
	    //Rentiere sind da
	    //Mutex fr die globale Variable freigeben
	    up(SEM_MUTEX);

	    printf("All reindeers have arrived.\n");

	    //ok, jetzt die Rentiere anspannen
            for(i=0; i<ALL_REINDEERS; i++)
	    {
		 duration = random()%3;
		 sleep(duration);
		 up(SEM_WORK);
		 printf("Santa yokes reindeer %i.\n", i);
	    }

	    //Weihnachten, Santa bringt die Geschenke
            printf("\n\nIt's christmas time!!!\n\n");
	    //Geschenke verteilen dauert ein bisschen
	    sleep(8);
	    
	    //ok, die Rentiere wieder abspannen
	    for(i=0; i<ALL_REINDEERS; i++)
	    {
		 duration = random()%3;
		 sleep(duration);
		 up(SEM_WORK_DONE);
		 printf("Santa unyokes reindeer %i.\n", i);
	    }

	    //jetzt sagen sie noch alle tschss
	    //ok, die Rentiere wieder abspannen
	    for(i=0; i<ALL_REINDEERS; i++)
	    {
		 duration = random()%5;
		 sleep(duration);
		 down(SEM_BYE);
		 printf("Reindeer %i says bye.\n", i);
	    }
	}
	else if(shmPtr->numberOfElfes == ALL_ELFES)
        {
	    //Nach den Rentieren sind die Elfen dran
	    //sind drei Elfen da?
	    //Mutex fr die globale Variable freigeben
	    up(SEM_MUTEX);
	    printf("Three elfes need help.\n");

	    duration = random()%5;
	    sleep(duration);

	    //ok, jetzt die Elfen rufen
            for(i=0; i<ALL_ELFES; i++)
	    {
		 up(SEM_GO_SANTA);
		 printf("Elf %i can go to santa.\n", i);
	    }

	    printf("Santa is giving advises.\n");
	    
	    //das kann natrlich dauern, wenn der Weihnachtsmann
	    //erst mal mit guten Ratschlgen anfngt
	    duration = random()%8;
	    sleep(duration);

	    //die Elfen werden wieder weggeschickt
	    for(i=0; i<ALL_ELFES; i++)
	    {
		 up(SEM_ADVISE_GIVEN);
		 printf("Elf %i can leave.\n", i);
	    }

	    //die Elfen bedanken sich
	    for(i=0; i<ALL_ELFES; i++)
	    {
		 duration = random()%5;
		 sleep(duration);
		 down(SEM_THANKS);
	         printf("Elf %i thanks santa.\n", i);
	    }
	}
	else
	{   
	    //Mutex fr die globale Variable freigeben
	    up(SEM_MUTEX);
	}
    }
}


